from __future__ import annotations

import io
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path

import polars as pl


@dataclass(frozen=True)
class Table:
    index: str

    @cached_property
    def fname(self) -> Path:
        return Path(__file__).parent / f"data/{self.index}.txt"

    @cached_property
    def raw(self) -> str:
        with open(self.fname, encoding="utf-8") as f:
            return f.read()

    @cached_property
    def header(self) -> list[str]:
        return self.raw.split("\n", 1)[0].split("\t", 1)

    @cached_property
    def name(self) -> str:
        return self.header[0]

    @cached_property
    def formula(self) -> str:
        return self.header[1]

    @cached_property
    def body(self) -> str:
        return self.raw.split("\n", 1)[1]

    @cached_property
    def df(self) -> pl.DataFrame:
        is_note = pl.col("delta-f H").str.contains(
            "<-->|TRANSITION|PRESSURE|FUGACITY|Cp|UNDEFINED"
        )
        return (
            pl.read_csv(
                io.StringIO(self.body),
                has_header=True,
                separator="\t",
                schema={
                    "T(K)": pl.Float64,
                    "Cp": pl.Float64,
                    "S": pl.Float64,
                    "-[G-H(Tr)]/T": pl.Float64,
                    "H-H(Tr)": pl.Float64,
                    "delta-f H": pl.String,
                    "delta-f G": pl.Float64,
                    "log Kf": pl.Float64,
                },
                truncate_ragged_lines=True,
            )
            .sort("T(K)")
            .with_columns(
                pl.when(is_note)
                .then(None)
                .otherwise(pl.col("delta-f H"))
                .cast(pl.Float64)
                .alias("delta-f H"),
                Note=pl.when(is_note).then(pl.col("delta-f H")),
            )
        )



from __future__ import annotations

import json

import polars as pl
import pytest

import janaf


def get_index() -> list[str]:
    with open("src/janaf/janaf.json", "rb") as f:
        return json.load(f)["index"]


@pytest.mark.parametrize("index", get_index())
def test_columns(index: str):
    table = janaf.Table(index)
    assert table.df.columns == [
        "T(K)",
        "Cp",
        "S",
        "-[G-H(Tr)]/T",
        "H-H(Tr)",
        "delta-f H",
        "delta-f G",
        "log Kf",
        "Note",
    ]
    assert table.df.dtypes == [
        pl.Float64,
        pl.Float64,
        pl.Float64,
        pl.Float64,
        pl.Float64,
        pl.Float64,
        pl.Float64,
        pl.Float64,
        pl.String,
    ]



import pytest

import janaf


def test_db():
    assert janaf.db.shape == (1_796, 5)


def test_search_formula():
    t = janaf.search(formula="CO$")
    assert t.name == "Carbon Monoxide (CO)"
    assert t.formula == "C1O1(g)"


def test_search_name():
    t = janaf.search(name="Water, 1 Bar")
    assert t.name == "Water, 1 Bar (H2O)"
    assert t.formula == "H2O1(l,g)"


def test_search_phase():
    t = janaf.search(formula="Fe$", phase="ref")
    assert t.name == "Iron (Fe)"
    assert t.formula == "Fe1(ref)"


def test_notunique():
    with pytest.raises(janaf.index.NotUnique):
        _ = janaf.search(formula="Fe$")



import os
import subprocess
import sys
from pathlib import Path

import pytest


def get_example_scripts():
    return [
        pytest.param(fp, id=fp.name)
        for fp in Path(__file__).parents[1].glob("example/**/*.py")
    ]


@pytest.mark.parametrize(
    "fp",
    get_example_scripts(),
)
def test_example(fp: Path):
    env = os.environ.copy()
    env["CI"] = "true"
    ret = subprocess.run(
        [sys.executable, fp.absolute()],
        env=env,
    )
    assert ret.returncode == 0



from __future__ import annotations

from typing import TYPE_CHECKING

from janaf import index, table
from janaf._version import __version__
from janaf.index import search
from janaf.table import Table

if TYPE_CHECKING:
    from typing import Any

    import polars as pl


def __getattr__(name: str) -> Any:
    from janaf.index import db

    if name == "db":
        return db()
    msg = f"module '{__name__}' has no attribute '{name}'"
    raise AttributeError(msg)


db: pl.DataFrame

__all__ = [
    "__version__",
    "db",
    "index",
    "search",
    "table",
    "Table",
]



from __future__ import annotations

import asyncio
import json
from pathlib import Path

import aiofiles
import aiohttp

root = Path(__file__).parent.parent.absolute()
path_json = root / "src/janaf/janaf.json"


def url(index: str) -> str:
    return f"https://janaf.nist.gov/tables/{index}.txt"


def dst(index: str) -> Path:
    return root / f"src/janaf/data/{index}.txt"


async def fetch(session: aiohttp.ClientSession, url: str, dst: Path) -> None:
    async with session.get(url) as res:
        data = await res.text("utf-8")
        async with aiofiles.open(dst, "w", encoding="utf-8") as f:
            await f.write(data)


async def main() -> None:
    async with aiofiles.open(path_json, encoding="utf-8") as fp:
        indexes = json.load(fp.buffer)["index"]

    async with aiohttp.ClientSession() as session:
        await asyncio.gather(
            *[fetch(session, url(index), dst(index)) for index in indexes],
        )


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

import asyncio
from pathlib import Path

import aiofiles
import aiohttp

url = "https://janaf.nist.gov/dat/janaf.json"
root = Path(__file__).parent.parent.absolute()
dst = root / "src/janaf/janaf.json"


async def main() -> None:
    async with aiohttp.ClientSession() as session, session.get(url) as res:
        text = await res.text("utf-8")
        async with aiofiles.open(dst, "w", encoding="utf-8") as f:
            await f.write(text)


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

import csv
import io
import logging
import re
from pathlib import Path


def fix_missing_tab(text: str) -> str:
    # eg. C-083.txt
    # 288.5001548.088	111.062	118.520	-2.152	II <--> III
    #       ^^ : missing tab

    if re.search(r"\.\d+\.", text.split("\n", 1)[1]) is None:
        return text

    return re.sub(r"(\.\d{3})(\d)", r"\1\t\2", text)


def trim_comment(text: str) -> str:
    # eg. S-021.txt
    # 1000	180.064	638.302	516.162	122.140	-392.026	52.900	-2.763
    # H
    # 1100	180.552	655.488	528.059	140.172	-388.966	97.243	-4.618

    title, header, body = text.split("\n", 2)

    new_body = (
        "\n".join(line for line in body.splitlines() if line[None:1] != "H") + "\n"
    )
    return f"{title}\n{header}\n{new_body}"


def combine_note(text: str) -> str:
    # eg. Ba-001.txt
    # 582.530	54.400	87.232	68.471	10.929	ALPHA <--> BETA
    # +
    # 582.530	32.468	87.232	68.471	10.929	TRANSITION

    title, header, body = text.split("\n", 2)

    if "+" not in body:
        return text

    lines = body.splitlines()
    for i in range(len(lines)):
        if "+" != lines[i][0]:
            continue

        if (r := re.search(r"\t([^\t]+)$", lines[i + 1])) is None:
            raise NotImplementedError
        target = r.group(1)
        lines[i - 1] = f"{lines[i-1]} {target}"
        lines[i] = ""
        lines[i + 1] = lines[i + 1].replace(target, "")

    new_body = "\n".join(x for x in lines if x != "").replace("\n\n", "\n") + "\n"
    return f"{title}\n{header}\n{new_body}"


def fix_missing_sign(text: str) -> str:
    # eg. Al-002.txt
    # 933.450	32.959	59.738	40.474	17.982	CRYSTAL <--> LIQUID
    # 1000	34.358	62.055	41.834	20.221	10.585 0.760  0.040
    #                                      ^             ^ : missing sign
    # 1100	36.722	65.438	43.827	23.772	-10.209	1.878	-0.089

    title, header, body = text.split("\n", 2)

    cell = [
        row
        for row in csv.reader(
            body.splitlines(),
            delimiter="\t",
            quoting=csv.QUOTE_NONE,
        )
    ]

    for i in range(len(cell) - 1):
        if (
            cell[i][5][None:1]
            in {"", "-", "0", "1", "2", "3", "4", "5", "6", "7", "8", "9"}
            and " " in cell[i][5]
        ):
            assert len(cell[i]) == 6
            idx = i + 1
            while len(cell[idx]) != 8:
                idx += 1
            cell[i].extend(["", ""])
            h, g, kf = cell[i][5].split()
            cell[i][5] = f"{'-' if cell[idx][5][None:1] == '-' else ''}{h}"
            cell[i][6] = f"{'-' if cell[idx][6][None:1] == '-' else ''}{g}"
            cell[i][7] = f"{'-' if cell[idx][7][None:1] == '-' else ''}{kf}"

    with io.StringIO() as f:
        writer = csv.writer(
            f,
            delimiter="\t",
            quoting=csv.QUOTE_NONE,
            lineterminator="\n",
        )
        writer.writerows(cell)
        new_body = f.getvalue()

    return f"{title}\n{header}\n{new_body}"


def fix_inf(text: str) -> str:
    # eg. Al-001.txt
    # 0	0.	0.	INFINITE	-4.539	0.	0.	0.
    #           ^^^^^^^^
    title, header, body = text.split("\n", 2)

    new_body = body.replace("INFINITE", "+inf")
    return f"{title}\n{header}\n{new_body}"


def main() -> None:
    for fp in Path("src/janaf/data/").glob("*.txt"):
        orig = fp.read_text(encoding="ascii")
        text = orig
        try:
            text = fix_missing_tab(text)
            text = trim_comment(text)
            text = combine_note(text)
            text = fix_missing_sign(text)
            text = fix_inf(text)
        except:
            logging.error(fp)
            raise
        if text != orig:
            logging.info(f"fixed: {fp}")
            fp.write_text(text, encoding="ascii")


if __name__ == "__main__":
    logging.basicConfig(format="%(message)s", level=logging.INFO)
    main()



from __future__ import annotations

from functools import lru_cache
from pathlib import Path

import polars as pl

from janaf.table import Table


@lru_cache(maxsize=None)
def db() -> pl.DataFrame:
    return (
        pl.read_json(Path(__file__).parent / "janaf.json")
        .explode("*")
        .with_columns(
            pl.col("display")
            .str.extract_groups(
                r"^(?P<formula>[^,]+), (?P<name>.+) \((?P<phase>[^)]+)\)$"
            )
            .alias("_")
        )
        .unnest("_")
    )


class NotUnique(Exception):  # noqa: N818
    pass


def search(
    *,
    formula: str | None = None,
    name: str | None = None,
    phase: str | None = None,
) -> Table:
    """Search a compound

    Parameters
    ----------
    formula, optional
        Regex for `formula`, by default None
    name, optional
        Regex for `name`, by default None
    phase, optional
        Regex for `phase`, by default None

    Returns
    -------
        `janaf.Table`

    Raises
    ------
    NotUnique
        Occurs when search results are not unique
    """

    expr = pl.lit(True)
    if formula:
        expr &= pl.col("formula").str.contains(
            formula if formula[0] == "^" else f"^{formula}"
        )
    if name:
        expr &= pl.col("name").str.contains(name if name[0] == "^" else f"^{name}")
    if phase:
        expr &= pl.col("phase").str.contains(phase if phase[0] == "^" else f"^{phase}")
    result = db().filter(expr)

    if len(result) != 1:
        raise NotUnique("\n" + str(result))
    return Table(index=result.select("index").item())



# /// script
# requires-python = ">=3.8"
# dependencies = [
#   "janaf",
#   "matplotlib",
#   "pandas >= 1.5.0",  # polars.DataFrame.to_pandas
#   "pyarrow >= 8.0.0",  # polars.DataFrame.to_pandas
# ]
# ///

import os

import matplotlib.pyplot as plt
import polars as pl
from matplotlib.axes import Axes

import janaf

t = janaf.search(name="Water, 1 Bar")

df = t.df.to_pandas(use_pyarrow_extension_array=True).set_index("T(K)")

fig, axes = plt.subplots(4, 2, sharex="col")
ax: Axes
for ax, col in zip(axes.flat, [c for c in df.columns if c not in {"Note"}]):
    df.plot(y=col, ax=ax, legend=False)
    ax.set_title(col)
    if col in {"Cp", "S", "-[G-H(Tr)]/T"}:
        ax.set_ylabel(f"{col} J/K/mol")
    if col in {"H-H(Tr)", "delta-f H", "delta-f G"}:
        ax.set_ylabel(f"{col} kJ/mol")

xs = (
    t.df.drop_nulls("Note").select(pl.col("T(K)").unique()).get_column("T(K)").to_list()
)
for ax in axes.flat:
    for x in xs:
        ax.axvline(x, color="k", linestyle=":")

fig.suptitle(f"{t.name} {t.formula}")
fig.tight_layout()
if not os.getenv("CI"):
    plt.show()
